from __future__ import annotations
from enum import Enum, auto
from typing import Any, Optional


class AST:
    statements: list[Statement]
    variables: set[Variable]

    def __init__(self, statements: list[Statement], variables: set[Variable]) -> None:
        self.statements = statements
        self.variables = variables

    def __repr__(self) -> str:
        stmt_str = "\n".join(map(str, self.statements))
        return f"<AST: {stmt_str}>"


class Statement:
    # Abstract class
    label: Optional[str]

    def __init__(self) -> None:
        raise NotImplementedError
    
    def get_label_string(self):
        label_str = "" 
        if self.label is not None: 
            label_str = f"({self.label}) "
        return label_str


class Assignment(Statement):
    variable: Variable
    value: Expression

    def __init__(self, var: Variable, val: Expression, label: Optional[str]) -> None:
        self.variable = var
        self.value = val
        self.label = label 

    def __str__(self) -> str:
        return f"{self.get_label_string()}{self.variable.id} := {self.value}"

    def __repr__(self) -> str:
        return f"<Assignment: {self.variable.id} = {self.value}>"


class Guard(Statement):
    expression: Expression

    def __init__(self, expr: Expression, label: Optional[str]) -> None:
        self.expression = expr
        self.label = label 

    def __str__(self) -> str:
        return f"{self.get_label_string()}[{self.expression}]"

    def __repr__(self) -> str:
        return f"<Guard: {self.expression}>"


class WhileLoop(Statement):
    condition: Expression
    statements: list[Statement]

    def __init__(self, condition: Expression, statements: list[Statement], label: Optional[str]) -> None:
        self.condition = condition
        self.statements = statements
        self.label = label

    def __str__(self) -> str:
        return f"{self.get_label_string()}while ({self.condition}) {self.statements}"

    def __repr__(self) -> str: 
        return f"<WhileLoop: while ({self.condition}) {self.statements}>"


class Branch(Statement):
    condition: Expression
    branch1: list[Statement]
    branch2: list[Statement]

    def __init__(self, condition: Expression, br1: list[Statement], br2: list[Statement], label: Optional[str]) -> None:
        self.condition = condition
        self.branch1 = br1
        self.branch2 = br2
        self.label = label

    def __str__(self) -> str:
        return f"{self.get_label_string()}if ({self.condition}) {self.branch1} else {self.branch2}"

    def __repr__(self) -> str:
        return f"<Branch: if ({self.condition}) {self.branch1} else {self.branch2}>"


class Expression:
    # Abstract class
    def __init__(self) -> None:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {str(self)}>"

    def __str__(self) -> str:
        raise NotImplementedError
    
    def __eq__(self, __o: object) -> bool:
        return type(__o) == type(self) and \
            repr(__o) == repr(self)
    
    def __hash__(self) -> int:
        return hash((type(self), repr(self)))


class BinaryExpression(Expression):
    operator: BinOp
    lhs: Expression
    rhs: Expression

    def __init__(self, lhs: Expression, operator: BinOp, rhs: Expression) -> None:
        self.lhs = lhs
        self.operator = operator
        self.rhs = rhs

    def __str__(self) -> str:
        if self.operator == BinOp.Plus:
            op_str = "+"
        if self.operator == BinOp.Minus:
            op_str = "-"
        if self.operator == BinOp.Times:
            op_str = "*"
        if self.operator == BinOp.Div:
            op_str = "/"
        if self.operator == BinOp.Or:
            op_str = "||"
        if self.operator == BinOp.And:
            op_str = "&&"
        if self.operator == BinOp.LessThan:
            op_str = "<"
        if self.operator == BinOp.LessThanEquals:
            op_str = "<="
        if self.operator == BinOp.GreaterThan:
            op_str = ">"
        if self.operator == BinOp.GreaterThanEquals:
            op_str = ">="
        if self.operator == BinOp.Equals:
            op_str = "=="
        if self.operator == BinOp.NotEquals:
            op_str = "!="
        left_str = f"({self.lhs})" if not isinstance(
            self.lhs, (ConstantExpression, Variable)) else str(self.lhs)
        right_str = f"({self.rhs})" if not isinstance(
            self.rhs, (ConstantExpression, Variable)) else str(self.rhs)
        return f"{left_str} {op_str} {right_str}"


class UnaryExpression(Expression):
    operator: UnOp
    expression: Expression

    def __init__(self, operator: UnOp, expression: Expression) -> None:
        self.operator = operator
        self.expression = expression

    def __str__(self) -> str:
        if self.operator == UnOp.Minus:
            op_str = "-"
        elif self.operator == UnOp.Not:
            op_str = "!"
        expr_str = f"({self.expression})" if not isinstance(
            self.expression, (ConstantExpression, Variable)) else str(self.expression)
        return f"{op_str}{expr_str}"


class ConstantExpression(Expression):
    # Abstract class
    value: Any

    def __init__(self) -> None:
        raise NotImplementedError

    def __str__(self) -> str:
        return str(self.value)


class IntegerConstant(ConstantExpression):
    def __init__(self, v: int) -> None:
        assert (isinstance(v, int))
        self.value = v


class BooleanConstant(ConstantExpression):
    def __init__(self, v: bool) -> None:
        assert (isinstance(v, bool))
        self.value = v


class Variable(Expression):
    # Note: I am overloading this class by using it for variable constant
    # expressions as well as for variables' other occurences.
    id: str

    def __init__(self, v: str) -> None:
        assert (isinstance(v, str))
        self.id = v

    def __repr__(self) -> str:
        return f"<Variable: {str(self)}>"

    def __str__(self) -> str:
        return self.id
    
    def __lt__(self, o):
        if isinstance(o, Variable): 
            return self.id < o.id
        
    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, Variable) and self.id == __o.id
    
    def __hash__(self) -> int:
        return hash(("Variable", self.id))


class BinOp(Enum):

    # int x int -> int
    Plus = auto()
    Minus = auto()
    Times = auto()
    Div = auto()

    # bool x bool -> bool
    Or = auto()
    And = auto()

    # int x int -> bool
    LessThan = auto()
    LessThanEquals = auto()
    GreaterThan = auto()
    GreaterThanEquals = auto()

    # any x any -> bool
    Equals = auto()
    NotEquals = auto()


class UnOp(Enum):

    Minus = auto()

    Not = auto()
