
from __future__ import annotations

from typing import Any, Optional, Set, Dict
from lang.ast import Statement, Expression, BinaryExpression, BooleanConstant, IntegerConstant, UnaryExpression, Assignment, Guard, Variable
from lang.cfg import CFG, Entry, Exit, InnerNode, Node


def getVarsUsedInStatement(stmt: Statement) -> Set[Variable]:
    if isinstance(stmt, Guard):
        return getVarsUsedInExpression(stmt.expression)
    elif isinstance(stmt, Assignment):
        return getVarsUsedInExpression(stmt.value)
    else:
        raise Exception(
            f"Statement {stmt} not supported for function getStatementUsedVars.")


def getVarsUsedInExpression(expr: Expression) -> Set[Variable]:
    if isinstance(expr, Variable):
        return {expr}
    elif isinstance(expr, BinaryExpression):
        return getVarsUsedInExpression(expr.lhs).union(getVarsUsedInExpression(expr.rhs))
    elif isinstance(expr, UnaryExpression):
        return getVarsUsedInExpression(expr.expression)
    elif isinstance(expr, (IntegerConstant, BooleanConstant)):
        return set()
    else:
        raise Exception(
            f"Expression {expr} not supported for function getExpressionUsedVars.")


def getValueByLabel(label: str, d: Dict[Node, Any]) -> Optional[Any]:
    for k in d.keys():
        if isinstance(k, InnerNode) and k.statement.label == label:
            return d[k]
    return None


def getEntryValue(d: Dict[Node, Any]) -> Optional[Any]:
    for k in d.keys():
        if isinstance(k, Entry):
            return d[k]
    return None


def getExitValue(d: Dict[Node, Any]) -> Optional[Any]:
    for k in d.keys():
        if isinstance(k, Exit):
            return d[k]
    return None


def getNodeByLabel(label: str, nodes: Set[Node]) -> Optional[Node]:
    for node in nodes:
        if isinstance(node, InnerNode):
            if node.statement.label == label:
                return node
    return None

def getEntryNode(nodes: Set[Node]) -> Optional[Node]:
    for node in nodes:
        if isinstance(node, Entry):
                return node
    return None

def getExitNode(nodes: Set[Node]) -> Optional[Node]:
    for node in nodes:
        if isinstance(node, Exit):
                return node
    return None

def getExpressionsInCFG(cfg: CFG, include_atomic=False) -> Set[Expression]:
    expressions = set()
    for node in cfg.nodes:
        if isinstance(node, InnerNode):
            expressions.update(getExpressionsInStatement(node.statement, include_atomic))
    return expressions


def getExpressionsInStatement(stmt: Statement, include_atomic=False) -> Set[Expression]:
    if isinstance(stmt, Guard):
        return getExpressionsInExpression(stmt.expression, include_atomic)
    elif isinstance(stmt, Assignment):
        return getExpressionsInExpression(stmt.value, include_atomic)
    else:
        raise Exception(
            f"Statement {stmt} with type {type(stmt)} not supported in getExpressionsInStatement.")


def getExpressionsInExpression(expr: Expression, include_atomic=False) -> Set[Expression]:
    if isinstance(expr, Variable):
        if include_atomic: 
            return { expr }
        return set()
    elif isinstance(expr, BinaryExpression):
        return {expr}.union(getExpressionsInExpression(expr.lhs, include_atomic).union(getExpressionsInExpression(expr.rhs, include_atomic)))
    elif isinstance(expr, UnaryExpression):
        return {expr}.union(getExpressionsInExpression(expr.expression, include_atomic))
    elif isinstance(expr, (IntegerConstant, BooleanConstant)):
        if include_atomic:
            return { expr }
        return set()
    else:
        raise Exception(
            f"Expression {expr} with type {type(expr)} not supported in getExpressionsInExpression.")
