from lang import stringToCFG
from lang.ast import Variable

# A note on using strings for checking CFG correctness,
# such as comparing a string against the output of str(cfg),
# cfg.str_next(), or cfg.str_prev():

# If there are duplicate statements in the program,
# they will be displayed in a nondeterministic order.
# So, one CFG may produce two different string representations.

# This may result in non-deterministic test failures.
# So, aim to write tests that have deterministic string reps
# or avoid using strings during testing and write tests that
# check order-insensitive features of the CFG, e.g. sets of nodes.


def test_cfg_simple():
    s = """
    x = 1; 
    y = 2;
    """
    cfg = stringToCFG(s)
    next = """\
y := 2               -> ['Exit']
x := 1               -> ['y := 2']
Exit                 -> []
Entry                -> ['x := 1']\
"""
    prev = """\
y := 2               -> ['x := 1']
x := 1               -> ['Entry']
Exit                 -> ['y := 2']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable("y")})


def test_cfg_branch():
    s = """
    if (x == 10) {
        y = 2;
    } else {
        y = 10; 
    }
    """
    cfg = stringToCFG(s)
    next = """\
y := 2               -> ['Exit']
y := 10              -> ['Exit']
[x == 10]            -> ['y := 2', 'y := 10']
Exit                 -> []
Entry                -> ['[x == 10]']\
"""
    prev = """\
y := 2               -> ['[x == 10]']
y := 10              -> ['[x == 10]']
[x == 10]            -> ['Entry']
Exit                 -> ['y := 2', 'y := 10']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable("y")})


def test_cfg_while():
    s = """
    while (z < 10) {
        x = 1; 
        y = 10;
    }
    """
    cfg = stringToCFG(s)
    next = """\
y := 10              -> ['[z < 10]']
x := 1               -> ['y := 10']
[z < 10]             -> ['x := 1', 'Exit']
Exit                 -> []
Entry                -> ['[z < 10]']\
"""
    prev = """\
y := 10              -> ['x := 1']
x := 1               -> ['[z < 10]']
[z < 10]             -> ['y := 10', 'Entry']
Exit                 -> ['[z < 10]']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable("y"), Variable("z")})


def test_cfg_branch_in_while():
    s = """
        while (z < 10) {
        if (x==1) {
            y = 10;
        } else {
            k = 1;
        }
    }
    """
    cfg = stringToCFG(s)
    next = """\
y := 10              -> ['[z < 10]']
k := 1               -> ['[z < 10]']
[z < 10]             -> ['[x == 1]', 'Exit']
[x == 1]             -> ['y := 10', 'k := 1']
Exit                 -> []
Entry                -> ['[z < 10]']\
"""
    prev = """\
y := 10              -> ['[x == 1]']
k := 1               -> ['[x == 1]']
[z < 10]             -> ['y := 10', 'k := 1', 'Entry']
[x == 1]             -> ['[z < 10]']
Exit                 -> ['[z < 10]']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable(
        "y"), Variable("z"), Variable("k")})


def test_cfg_while_in_branch1():
    s = """
    if (x < (y + 10)) {
        while (x < y) {
            x = x + 1;
        }
    } else {

    }
    """
    cfg = stringToCFG(s)
    next = """\
x := x + 1           -> ['[x < y]']
[x < y]              -> ['x := x + 1', 'Exit']
[x < (y + 10)]       -> ['[x < y]', 'Exit']
Exit                 -> []
Entry                -> ['[x < (y + 10)]']\
"""
    prev = """\
x := x + 1           -> ['[x < y]']
[x < y]              -> ['x := x + 1', '[x < (y + 10)]']
[x < (y + 10)]       -> ['Entry']
Exit                 -> ['[x < y]', '[x < (y + 10)]']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable("y")})


def test_cfg_while_in_branch2():
    s = """
    if (x < (y + 10)) {
        
    } else {
        while (x < y) {
            x = x + 1;
        }
    }
    """
    cfg = stringToCFG(s)
    next = """\
x := x + 1           -> ['[x < y]']
[x < y]              -> ['x := x + 1', 'Exit']
[x < (y + 10)]       -> ['[x < y]', 'Exit']
Exit                 -> []
Entry                -> ['[x < (y + 10)]']\
"""
    prev = """\
x := x + 1           -> ['[x < y]']
[x < y]              -> ['x := x + 1', '[x < (y + 10)]']
[x < (y + 10)]       -> ['Entry']
Exit                 -> ['[x < y]', '[x < (y + 10)]']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable("y")})


def test_cfg_same_while_in_both_branch():
    s = """
    if (x < (y + 10)) {
        while (x < y) {
            x = x + 1;
        }
    } else {
        while (x < y) {
            x = x + 1;
        }
    }
    """
    cfg = stringToCFG(s)
    next = """\
x := x + 1           -> ['[x < y]']
x := x + 1           -> ['[x < y]']
[x < y]              -> ['x := x + 1', 'Exit']
[x < y]              -> ['x := x + 1', 'Exit']
[x < (y + 10)]       -> ['[x < y]', '[x < y]']
Exit                 -> []
Entry                -> ['[x < (y + 10)]']\
"""
    prev = """\
x := x + 1           -> ['[x < y]']
x := x + 1           -> ['[x < y]']
[x < y]              -> ['x := x + 1', '[x < (y + 10)]']
[x < y]              -> ['x := x + 1', '[x < (y + 10)]']
[x < (y + 10)]       -> ['Entry']
Exit                 -> ['[x < y]', '[x < y]']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable("y")})


def test_cfg_diff_while_in_each_branch():
    s = """
    if (x < (y + 10)) {
        while (x < y) {
            x = x + 1;
        }
    } else {
        while (x >= z) {
            k = x + 1;
        }
    }
    """
    cfg = stringToCFG(s)
    next = """\
x := x + 1           -> ['[x < y]']
k := x + 1           -> ['[x >= z]']
[x >= z]             -> ['k := x + 1', 'Exit']
[x < y]              -> ['x := x + 1', 'Exit']
[x < (y + 10)]       -> ['[x >= z]', '[x < y]']
Exit                 -> []
Entry                -> ['[x < (y + 10)]']\
"""
    prev = """\
x := x + 1           -> ['[x < y]']
k := x + 1           -> ['[x >= z]']
[x >= z]             -> ['k := x + 1', '[x < (y + 10)]']
[x < y]              -> ['x := x + 1', '[x < (y + 10)]']
[x < (y + 10)]       -> ['Entry']
Exit                 -> ['[x >= z]', '[x < y]']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable(
        "y"), Variable("z"), Variable("k")})


def test_cfg_multi():

    s = """
    x = 10; 
    if (x < (y + 10)) {
        while (x < y) {
            k = x + 1;
        }
        k = y; 
    } else {
        k = x;
        while (x < k) {
            x = x + 1;
        }
    }
    x = x + 2;"""
    cfg = stringToCFG(s)
    next = """\
x := x + 2           -> ['Exit']
x := x + 1           -> ['[x < k]']
x := 10              -> ['[x < (y + 10)]']
k := y               -> ['x := x + 2']
k := x               -> ['[x < k]']
k := x + 1           -> ['[x < y]']
[x < y]              -> ['k := y', 'k := x + 1']
[x < k]              -> ['x := x + 2', 'x := x + 1']
[x < (y + 10)]       -> ['k := x', '[x < y]']
Exit                 -> []
Entry                -> ['x := 10']\
"""
    prev = """\
x := x + 2           -> ['k := y', '[x < k]']
x := x + 1           -> ['[x < k]']
x := 10              -> ['Entry']
k := y               -> ['[x < y]']
k := x               -> ['[x < (y + 10)]']
k := x + 1           -> ['[x < y]']
[x < y]              -> ['k := x + 1', '[x < (y + 10)]']
[x < k]              -> ['x := x + 1', 'k := x']
[x < (y + 10)]       -> ['x := 10']
Exit                 -> ['x := x + 2']
Entry                -> []\
"""
    assert (cfg.str_next() == next)
    assert (cfg.str_prev() == prev)
    assert (cfg.variables == {Variable("x"), Variable("y"), Variable("k")})
